// Software SPI module connections
sbit SoftSpi_SDI at RA2_bit;
sbit SoftSpi_SDO at RA0_bit;
sbit SoftSpi_CLK at RA1_bit;

sbit SoftSpi_SDI_Direction at TRISA2_bit;
sbit SoftSpi_SDO_Direction at TRISA0_bit;
sbit SoftSpi_CLK_Direction at TRISA1_bit;
// End Software SPI module connections

sbit CS at RA5_bit;
sbit CS_Direction at TRISA5_bit;

#define _LCD_FIRST_ROW          0x80     //Move cursor to the 1st row
#define _LCD_SECOND_ROW         0xC0     //Move cursor to the 2nd row
#define _LCD_THIRD_ROW          0x94     //Move cursor to the 3rd row
#define _LCD_FOURTH_ROW         0xD4     //Move cursor to the 4th row
#define _LCD_CLEAR              0x01     //Clear display
#define _LCD_RETURN_HOME        0x02     //Return cursor to home position, returns a shifted display to
                                         //its original position. Display data RAM is unaffected.
#define _LCD_CURSOR_OFF         0x0C     //Turn off cursor
#define _LCD_UNDERLINE_ON       0x0E     //Underline cursor on
#define _LCD_BLINK_CURSOR_ON    0x0F     //Blink cursor on
#define _LCD_MOVE_CURSOR_LEFT   0x10     //Move cursor left without changing display data RAM
#define _LCD_MOVE_CURSOR_RIGHT  0x14     //Move cursor right without changing display data RAM
#define _LCD_TURN_ON            0x0C     //Turn Lcd display on
#define _LCD_TURN_OFF           0x08     //Turn Lcd display off
#define _LCD_SHIFT_LEFT         0x18     //Shift display left without changing display data RAM
#define _LCD_SHIFT_RIGHT        0x1E     //Shift display right without changing display data RAM

#define  IODIRA   0x00
#define  IPOLA    0x01
#define  GPINTENA 0x02
#define  DEFVALA  0x03
#define  INTCONA  0x04
#define  IOCON    0x05
#define  GPPUA    0x06
#define  INTFA    0x07
#define  IINTCAPA 0x08
#define  GPIOA    0x09
#define  OLATA    0x0A

char str[23];

void SPI_LCD_Cmd(char out_char) {

    char hi_n, lo_n;
    char rs = 0x00;

    hi_n = out_char & 0xF0;
    lo_n = (out_char << 4) & 0xF0;

    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(hi_n | rs | 0x04 | 0x08);
    CS = 1;
    
    Delay_us(50);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(hi_n | rs | 0x00 | 0x08);
    CS = 1;
    
    Delay_us(100);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(lo_n | rs | 0x04 | 0x08);
    CS = 1;

    Delay_us(50);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(lo_n | rs | 0x00 | 0x08);
    CS = 1;
    
    if(out_char == 0x01)Delay_ms(2);
}

void SPI_LCD_Chr(char row, char column, char out_char) {

    char hi_n, lo_n;
    char rs = 0x01;

    switch(row){

        case 1:
        SPI_LCD_Cmd(0x80 + (column - 1));
        break;
        case 2:
        SPI_LCD_Cmd(0xC0 + (column - 1));
        break;
        case 3:
        SPI_LCD_Cmd(0x94 + (column - 1));
        break;
        case 4:
        SPI_LCD_Cmd(0xD4 + (column - 1));
        break;
    };

    hi_n = out_char & 0xF0;
    lo_n = (out_char << 4) & 0xF0;

    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(hi_n | rs | 0x04 | 0x08);
    CS = 1;
    
    Delay_us(50);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(hi_n | rs | 0x00 | 0x08);
    CS = 1;
    
    Delay_us(100);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(lo_n | rs | 0x04 | 0x08);
    CS = 1;
    
    Delay_us(50);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(lo_n | rs | 0x00 | 0x08);
    CS = 1;
}

void SPI_LCD_Chr_Cp(char out_char) {

    char hi_n, lo_n;
    char rs = 0x01;

    hi_n = out_char & 0xF0;
    lo_n = (out_char << 4) & 0xF0;

    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(hi_n | rs | 0x04 | 0x08);
    CS = 1;
    
    Delay_us(50);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(hi_n | rs | 0x00 | 0x08);
    CS = 1;

    Delay_us(100);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(lo_n | rs | 0x04 | 0x08);
    CS = 1;

    Delay_us(50);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(lo_n | rs | 0x00 | 0x08);
    CS = 1;
}


void SPI_LCD_Init() {

    char rs = 0x00;

    Delay_ms(30);

    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(0x30 | rs | 0x04 | 0x08);

    Delay_us(50);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(0x30 | rs | 0x00 | 0x08);
    CS = 1;


    Delay_ms(10);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(0x30 | rs | 0x04 | 0x08);
    CS = 1;

    Delay_us(50);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(0x30 | rs | 0x00 | 0x08);
    CS = 1;


    Delay_ms(10);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(0x30 | rs | 0x04 | 0x08);
    CS = 1;

    Delay_us(50);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(0x30 | rs | 0x00 | 0x08);
    CS = 1;


    Delay_ms(10);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(0x20 | rs | 0x04 | 0x08);
    CS = 1;

    Delay_us(50);
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPIOA);
    Soft_Spi_Write(0x20 | rs | 0x00 | 0x08);
    CS = 1;

    Delay_ms(10);

    SPI_LCD_Cmd(0x28);
    SPI_LCD_Cmd(0x06);
}

void SPI_LCD_Out(char row, char col, char *text) {
    while(*text)
         SPI_LCD_Chr(row, col++, *text++);
}

void SPI_LCD_Out_Cp(char *text) {
    while(*text)
         SPI_LCD_Chr_Cp(*text++);
}

void main() {

    CM1CON0 = 0x07;
    ANSELA = 0x00;
    TRISA= 0x08;
    PORTA = 0x00;
       
    CS_Direction = 0;
    CS = 1;

    Soft_Spi_Init();

    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(IOCON);
    Soft_Spi_Write(0x27);
    CS = 1;
    
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(IODIRA);
    Soft_Spi_Write(0x00);
    CS = 1;

    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(IPOLA);
    Soft_Spi_Write(0x00);
    CS = 1;
    
    CS = 0;
    Soft_Spi_Write(0x40);
    Soft_Spi_Write(GPPUA);
    Soft_Spi_Write(0xFF);
    CS = 1;

    SPI_LCD_Init();
    SPI_LCD_Cmd(_LCD_CURSOR_OFF);
    SPI_LCD_Cmd(_LCD_CLEAR);

    SPI_LCD_Out(1,1,"MCP23S09 Based");
    SPI_LCD_Out(2,1,"Soft Spi LCD");
    SPI_LCD_Out(3,1,"16x2, 20x2, 20x4");
    SPI_LCD_Out(4,1,"PIC12F1840");

    while(1) {
     
    }
}